#![no_std]
#![no_main]
#![feature(type_alias_impl_trait)]
#![feature(never_type)]
#![feature(async_fn_in_trait)]

#[cfg(not(feature = "serial"))]
mod mouse_hid;

#[cfg(feature = "serial")]
mod usb_serial;

mod mouse_sensor;

use embassy_executor::Spawner;
use embassy_futures::join::join;
use embassy_rp::{
    peripherals::USB,
    spi::{CsPin, Instance},
};

use embassy_time::{Duration, Timer};
use mouse_sensor::PAW3212DB_TJDT;
#[cfg(not(feature = "serial"))]
use usbd_hid::descriptor::MouseReport;

#[cfg(not(feature = "serial"))]
use crate::{mouse_hid::mouse_config::MouseConfig, mouse_hid::MouseHID};

#[cfg(feature = "serial")]
use crate::usb_serial::{Serial, SerialConfig};

use crate::usb_serial::Sender;

use crate::mouse_sensor::{MouseSensor, MouseSensorPins};

use {defmt_rtt as _, panic_probe as _};

#[embassy_executor::main]
async fn main(_spawner: Spawner) {
    let p = embassy_rp::init(Default::default());

    // let mouse_sensor = MouseSensor::new(
    //     p.PIN_25,
    //     MouseSensorPins::new(p.SPI0, p.PIN_2, p.PIN_3, p.PIN_4, p.PIN_5),
    //     |_| (),
    // )
    // .await;

    // #[cfg(not(feature = "serial"))]
    // mouse_hid(p.USB, mouse_sensor).await;

    // #[cfg(feature = "serial")]
    // serial_hid(p.USB, mouse_sensor).await;

    #[cfg(feature = "serial")]
    {
        let serial = Serial::new(p.USB, SerialConfig::new());

        let usb_future = Serial::run_usb().await;

        // let mut mouse_sensor = MouseSensor::new(
        //     p.PIN_25,
        //     MouseSensorPins::new(p.SPI0, p.PIN_2, p.PIN_3, p.PIN_4, p.PIN_5),
        //     serial,
        // )
        // .await;

        let sensor_pins = MouseSensorPins::new(p.SPI0, p.PIN_2, p.PIN_3, p.PIN_4, p.PIN_5);

        let mut spi_device = PAW3212DB_TJDT::new(
            sensor_pins.spi,
            sensor_pins.clk,
            sensor_pins.mosi,
            sensor_pins.miso,
            sensor_pins.cs,
            serial,
        )
        .await;

        // Do stuff with the class!
        let hid_future = async {
            loop {
                // if let Some((x, y)) = mouse_sensor.read_movement().await {
                //     // let report = MouseReport {
                //     //     buttons: 0,
                //     //     x,
                //     //     y,
                //     //     wheel: 0,
                //     //     pan: 0,
                //     // };

                //     // serial.send_number(x, 10).await;
                //     // serial.send_number(y, 10).await;
                // }

                let id_1 = spi_device.product_id_1().await;
                spi_device.sender.send_msg("read pid1:\n").await;
                spi_device.sender.send_number(id_1, 10).await;
                spi_device.sender.send_msg("\n").await;
                Timer::after(Duration::from_secs(1)).await;
            }
        };

        join(hid_future, usb_future).await;
    }
}

#[cfg(not(feature = "serial"))]
async fn mouse_hid<'d, L, T, CSP>(usb: USB, mut mouse_sensor: MouseSensor<'d, L, T, CSP>)
where
    L: FnMut(&str) + 'd,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    let mut mouse_hid = MouseHID::new(usb, MouseConfig::new());

    let usb_future = MouseHID::run_usb().await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                let report = MouseReport {
                    buttons: 0,
                    x,
                    y,
                    wheel: 0,
                    pan: 0,
                };

                #[cfg(not(feature = "serial"))]
                mouse_hid.mouse_report(report).await;
            }
        }
    };

    join(hid_future, usb_future).await;
}

#[cfg(feature = "serial")]
async fn serial_hid<'d, S, T, CSP>(usb: USB, mut mouse_sensor: MouseSensor<'d, S, T, CSP>)
where
    S: Sender,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    let mut serial = Serial::new(usb, SerialConfig::new());

    let usb_future = Serial::run_usb().await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                // let report = MouseReport {
                //     buttons: 0,
                //     x,
                //     y,
                //     wheel: 0,
                //     pan: 0,
                // };

                serial.send_number(x, 10).await;
                serial.send_number(y, 10).await;
            }
        }
    };

    join(hid_future, usb_future).await;
}
